from __future__ import absolute_import, unicode_literals
import os
from celery import Celery

import platform

if platform.system().lower() == "windows":
    os.environ.setdefault("FORKED_BY_MULTIPROCESSING", "1")

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "arches_modular_reports.settings")
app = Celery("arches_modular_reports")
app.config_from_object("django.conf:settings", namespace="CELERY")
app.autodiscover_tasks()



import re
from django_hosts import patterns, host

host_patterns = patterns(
    "",
    host(
        re.sub(r"_", r"-", r"arches_modular_reports"),
        "arches_modular_reports.urls",
        name="arches_modular_reports",
    ),
)



# Generated by Django 5.1.2 on 2024-11-01 18:44

import django.db.models.deletion
from django.db import migrations, models

import arches_modular_reports.utils
from arches.app.models.system_settings import settings


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("models", "11499_add_editlog_resourceinstance_idx"),
    ]

    operations = [
        migrations.CreateModel(
            name="ReportConfig",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "config",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=arches_modular_reports.utils.PrettyJSONEncoder,
                    ),
                ),
                (
                    "graph",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="report",
                        to="models.graphmodel",
                        limit_choices_to=models.Q(isresource=True)
                        & ~models.Q(pk=settings.SYSTEM_SETTINGS_RESOURCE_MODEL_ID),
                    ),
                ),
            ],
            options={
                "db_table": "arches_modular_report_config",
                "managed": True,
            },
        ),
    ]



"""
ARCHES - a program developed to inventory and manage immovable cultural heritage.
Copyright (C) 2013 J. Paul Getty Trust and World Monuments Fund

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
import inspect

path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))

if path not in sys.path:
    sys.path.append(path)

# reverting back to the old style of setting the DJANGO_SETTINGS_MODULE env variable
# refer to the following blog post under the heading "Leaking of process environment variables."
# http://blog.dscpl.com.au/2012/10/requests-running-in-wrong-django.html
os.environ["DJANGO_SETTINGS_MODULE"] = "arches_modular_reports.settings"

from django.core.wsgi import get_wsgi_application

application = get_wsgi_application()

from arches.app.models.system_settings import settings

settings.update_from_db()



# Generated by Django 5.2.3 on 2025-06-16 15:09

import arches_modular_reports.models
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("arches_modular_reports", "0004_handle_for_concept_label_subregions"),
    ]

    operations = [
        migrations.AlterField(
            model_name="reportconfig",
            name="graph",
            field=models.ForeignKey(
                limit_choices_to=arches_modular_reports.models.get_graph_choices,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="report",
                to="models.graphmodel",
            ),
        ),
    ]



import json
import operator
from functools import reduce
from uuid import UUID

from django.contrib.postgres.expressions import ArraySubquery
from django.db.models import (
    Case,
    Exists,
    F,
    Func,
    IntegerField,
    JSONField,
    OuterRef,
    Q,
    TextField,
    Value,
    When,
)
from django.db.models.expressions import CombinedExpression
from django.db.models.fields.json import KT
from django.db.models.functions import Cast, Concat, JSONObject
from django.urls import get_script_prefix, reverse
from django.utils.translation import gettext as _

from arches import VERSION as arches_version
from arches.app.models import models


class ArchesGetNodeDisplayValueV2(Func):
    function = "__arches_get_node_display_value_v2"
    output_field = TextField()
    arity = 3


class ArchesGetValueId(Func):
    function = "__arches_get_valueid"
    output_field = TextField()
    arity = 3


class ArrayToString(Func):
    function = "ARRAY_TO_STRING"
    output_field = TextField()
    arity = 3


def get_link(datatype, value_id):
    if datatype in ["concept", "concept-list"]:
        return reverse("rdm", args=[value_id])
    elif datatype in ["resource-instance", "resource-instance-list"]:
        return reverse("resource_report", args=[value_id])
    elif datatype in ["url"]:
        return value_id
    elif datatype in ["reference"]:
        return value_id["uri"]
    return ""


def build_valueid_annotation(data, is_user_rdm_admin, user_language):
    datatype = data.get("datatype", "")
    display_value = data.get("display_value")

    if datatype in ["concept", "resource-instance"]:
        value_ids = data.get("value_ids")
        if value_ids and (
            (datatype == "concept" and is_user_rdm_admin)
            or datatype == "resource-instance"
        ):
            return {
                "display_value": [
                    {
                        "label": display_value,
                        "link": get_link(datatype, value_ids),
                    }
                ],
                "has_links": True,
            }
        return {"display_value": display_value, "has_links": False}

    elif datatype in ["concept-list", "resource-instance-list", "reference"]:
        display_values = json.loads(data.get("display_value", "[]"))
        value_ids = data.get("value_ids", "[]") if data.get("value_ids", "[]") else []

        if datatype in ["concept-list", "resource-instance-list"]:
            value_ids = json.loads(value_ids)

        annotations = []
        if datatype == "concept-list" and not is_user_rdm_admin:
            return {"display_value": ", ".join(display_values), "has_links": False}
        else:
            for val_id, disp_val in zip(value_ids, display_values):
                if val_id:
                    annotations.append(
                        {
                            "label": disp_val,
                            "link": get_link(datatype, val_id),
                        }
                    )
            return {"display_value": annotations, "has_links": True}

    elif datatype in ["url"]:
        value_ids = data.get("value_ids")
        if value_ids:
            return {
                "display_value": [
                    {
                        "label": (
                            display_value
                            if display_value != ""
                            else get_link(datatype, value_ids)
                        ),
                        "link": get_link(datatype, value_ids),
                    }
                ],
                "has_links": True,
            }
        return {"display_value": display_value, "has_links": False}

    elif datatype == "file-list":
        file_tile_value = data.get("file_tile_value")

        def get_localized_metadata(file_info, key):
            nonlocal user_language
            if not (localized_values := file_info.get(key)):
                return None
            if not (value_for_lang := localized_values.get(user_language)):
                # todo(polish): handle variation in language regions.
                return list(localized_values.values)[0]["value"]
            return value_for_lang["value"]

        if file_tile_value:
            return {
                "display_value": display_value,
                "file_data": [
                    {
                        "file_id": file_info.get("file_id"),
                        "url": file_info.get("url"),
                        "title": get_localized_metadata(file_info, "title"),
                        "attribution": get_localized_metadata(file_info, "attribution"),
                        "description": get_localized_metadata(file_info, "description"),
                        "altText": get_localized_metadata(file_info, "altText"),
                    }
                    for file_info in file_tile_value
                ],
                "is_file": True,
            }
        return {"display_value": display_value, "has_links": False}

    return {"display_value": display_value, "has_links": False}


def annotate_related_graph_nodes_with_widget_labels(
    additional_nodes, related_graph, request_language
):
    return (
        models.Node.objects.filter(alias__in=additional_nodes, graph=related_graph)
        .exclude(datatype__in=["semantic", "annotation", "geojson-feature-collection"])
        .annotate(
            widget_label_json=(
                models.CardXNodeXWidget.objects.filter(node=OuterRef("nodeid")).values(
                    "label"
                )[:1]
            )
        )
        .annotate(widget_label=KT(f"widget_label_json__{request_language}"))
    )


def annotate_node_values(
    node_aliases, resourceinstance_id, permitted_nodegroups, user_language, tile_limit
):
    tile_subquery = (
        models.TileModel.objects.filter(
            resourceinstance=resourceinstance_id,
            nodegroup_id=OuterRef("nodegroup_id"),
        )
        .annotate(
            json_object=JSONObject(
                display_value=ArchesGetNodeDisplayValueV2(
                    F("data"),
                    OuterRef("nodeid"),
                    Value(user_language),
                ),
                tile_value=CombinedExpression(
                    F("data"),
                    "->",
                    Cast(OuterRef("nodeid"), output_field=TextField()),
                    output_field=JSONField(),
                ),
            ),
        )
        .exclude(json_object__display_value="")
        .exclude(json_object__tile_value=None)
        # This will work on Django 5.1+
        # .distinct("json_object__display_value", "sortorder")
        .order_by("sortorder")
        .values("json_object")
    )
    if tile_limit:
        tile_subquery = tile_subquery[:tile_limit]

    return (
        models.Node.objects.filter(
            alias__in=node_aliases,
            graph__resourceinstance__pk=resourceinstance_id,
            nodegroup__in=permitted_nodegroups,
        )
        .exclude(datatype__in=["semantic", "annotation", "geojson-feature-collection"])
        .annotate(display_data=ArraySubquery(tile_subquery))
    )


def get_sorted_filtered_tiles(
    *,
    resourceinstanceid,
    nodegroup_alias,
    sort_node_id,
    direction,
    query,
    user_language,
    user,
    filters,
):
    # semantic, annotation, and geojson-feature-collection data types are
    # excluded in __arches_get_node_display_value
    nodes = models.Node.objects.filter(
        graph__resourceinstance=resourceinstanceid,
        nodegroup__node__alias=nodegroup_alias,
        nodegroup__in=user.userprofile.viewable_nodegroups,
    ).exclude(
        datatype__in={"semantic", "annotation", "geojson-feature-collection"},
    )
    if arches_version >= (8, 0):
        nodes = nodes.filter(source_identifier=None)

    if not nodes:
        return models.TileModel.objects.none()

    field_annotations = {}
    alias_annotations = {}
    tile_filters = Q()

    for node in nodes:
        field_key = f'field_{str(node.pk).replace("-", "_")}'

        display_value = ArchesGetNodeDisplayValueV2(
            F("data"), Value(str(node.pk)), Value(user_language)
        )

        value_ids = None
        tile_value = None
        if (
            node.datatype == "concept"
            or node.datatype == "concept-list"
            or node.datatype == "resource-instance"
            or node.datatype == "resource-instance-list"
            or node.datatype == "url"
        ):
            value_ids = ArchesGetValueId(
                F("data"), Value(node.pk), Value(user_language)
            )
        elif node.datatype == "file-list":
            tile_value = F(f"data__{node.pk}")
        elif node.datatype == "reference":
            value_ids = F(f"data__{node.pk}")

        field_annotations[field_key] = display_value
        alias_annotations[node.alias] = JSONObject(
            display_value=display_value,
            datatype=Value(node.datatype),
            value_ids=value_ids,
            file_tile_value=tile_value,
        )
        if filters:
            for filter in filters:
                if node.alias == filter["alias"]:
                    tile_filters &= Q(
                        **{
                            f"data__{node.pk}__{filter['field_lookup']}": filter[
                                "value"
                            ]
                        }
                    )

    # adds spaces between fields
    display_values_with_spaces = []
    for field in [F(field) for field in field_annotations.keys()]:
        display_values_with_spaces.append(field)
        display_values_with_spaces.append(Value(" "))

    tiles = (
        models.TileModel.objects.filter(
            resourceinstance_id=resourceinstanceid, nodegroup_id=nodes[0].nodegroup_id
        )
        .annotate(**field_annotations)
        .annotate(alias_annotations=JSONObject(**alias_annotations))
        .annotate(
            search_text=Concat(*display_values_with_spaces, output_field=TextField())
        )
        .filter(tile_filters & Q(search_text__icontains=query))
        .annotate(
            has_children=Exists(
                models.TileModel.objects.filter(parenttile=OuterRef("pk"))
            )
        )
    )

    if sort_node_id:
        sort_field_name = f'field_{sort_node_id.replace("-", "_")}'

        sort_priority = Case(
            When(**{f"{sort_field_name}__isnull": True}, then=Value(1)),
            When(**{f"{sort_field_name}": ""}, then=Value(1)),
            default=Value(0),
            output_field=IntegerField(),
        )

        if direction.lower().startswith("asc"):
            tiles = tiles.annotate(sort_priority=sort_priority).order_by(
                "sort_priority", F(sort_field_name).asc()
            )
        else:
            tiles = tiles.annotate(sort_priority=sort_priority).order_by(
                "-sort_priority", F(sort_field_name).desc()
            )
    else:
        # default sort order for consistent pagination
        tiles = tiles.order_by("sortorder")

    return tiles


def get_sorted_filtered_relations(
    *,
    resource,
    related_graph,
    nodes,
    permitted_nodegroups,
    sort_field,
    direction,
    query,
    request_language,
):
    if arches_version < (8, 0):
        resource_from_field = "resourceinstanceidfrom"
        resource_from_graph_field = "resourceinstancefrom_graphid"
        resource_to_field = "resourceinstanceidto"
        resource_to_graph_field = "resourceinstanceto_graphid"
        node_field = "nodeid"
    else:
        resource_from_field = "from_resource"
        resource_from_graph_field = "from_resource_graph"
        resource_to_field = "to_resource"
        resource_to_graph_field = "to_resource_graph"
        node_field = "node"

    def make_tile_annotations(node, direction):
        resource_field = resource_to_field if direction == "to" else resource_from_field
        tile_query = ArraySubquery(
            models.TileModel.objects.filter(
                resourceinstance=OuterRef(resource_field),
                nodegroup_id=node.nodegroup_id,
            )
            .exclude(**{f"data__{node.pk}__isnull": True})
            .annotate(
                display_value=ArchesGetNodeDisplayValueV2(
                    F("data"), Value(node.pk), Value(request_language)
                )
            )
            .order_by("sortorder")
            .values("display_value")
            .distinct()
        )
        return ArrayToString(
            tile_query,
            Value(", "),  # delimiter
            Value(_("None")),  # null replacement
        )

    def make_tile_instance_details_annotations(node, direction):
        resource_field = resource_to_field if direction == "to" else resource_from_field
        return ArraySubquery(
            models.TileModel.objects.filter(
                resourceinstance=OuterRef(resource_field),
                nodegroup_id=node.nodegroup_id,
            )
            .exclude(**{f"data__{node.pk}__isnull": True})
            .order_by("sortorder")
            .annotate(node_value=F(f"data__{node.pk}"))
            .values("node_value")
            .distinct()
        )

    data_annotations = {
        node.alias: Case(
            When(
                Q(**{resource_from_field: resource}),
                then=make_tile_annotations(node, "to"),
            ),
            When(
                Q(**{resource_to_field: resource}),
                then=make_tile_annotations(node, "from"),
            ),
        )
        for node in nodes
        if node.nodegroup_id in permitted_nodegroups
    }
    instance_details_annotations = {
        node.alias
        + "_instance_details": Case(
            When(
                Q(**{resource_from_field: resource}),
                then=make_tile_instance_details_annotations(node, "to"),
            ),
            When(
                Q(**{resource_to_field: resource}),
                then=make_tile_instance_details_annotations(node, "from"),
            ),
        )
        for node in nodes
        if node.datatype
        in {
            "concept",
            "concept-list",
            "resource-instance",
            "resource-instance-list",
            "url",
            "file-list",
        }
        and node.nodegroup_id in permitted_nodegroups
    }

    relations = (
        (
            models.ResourceXResource.objects.filter(
                Q(**{resource_from_field: resource}),
                Q(**{resource_to_graph_field: related_graph}),
                Q(**{f"{node_field}__nodegroup_id__in": permitted_nodegroups}),
            )
            | models.ResourceXResource.objects.filter(
                Q(**{resource_to_field: resource}),
                Q(**{resource_from_graph_field: related_graph}),
                Q(**{f"{node_field}__nodegroup_id__in": permitted_nodegroups}),
            )
        )
        .distinct()
        .annotate(
            relation_name_json=(
                models.CardXNodeXWidget.objects.filter(
                    node=OuterRef(node_field)
                ).values("label")[:1]
            )
        )
        # TODO: add fallback to system language? Below also.
        # https://github.com/archesproject/arches/issues/10028
        .annotate(**{"@relation_name": KT(f"relation_name_json__{request_language}")})
        .annotate(
            display_name_json=Case(
                When(
                    Q(**{resource_from_field: resource}),
                    then=F(f"{resource_to_field}__name"),
                ),
                When(
                    Q(**{resource_to_field: resource}),
                    then=F(f"{resource_from_field}__name"),
                ),
            )
        )
        .annotate(**{"@display_name": KT(f"display_name_json__{request_language}")})
        .annotate(**data_annotations)
        .annotate(**instance_details_annotations)
    )

    if query:
        # OR Q objects together to allow matching any annotation.
        all_filters = reduce(
            operator.or_,
            [
                Q(**{"@relation_name__icontains": query}),
                Q(**{"@display_name__icontains": query}),
                *[
                    Q(**{f"{annotation}__icontains": query})
                    for annotation in data_annotations
                ],
            ],
        )
        relations = relations.filter(all_filters)

    if direction.lower().startswith("asc"):
        relations = relations.order_by(F(sort_field).asc(nulls_last=True))
    else:
        relations = relations.order_by(F(sort_field).desc(nulls_last=True))

    return relations


def filter_hidden_nodes(
    list_or_dict_to_filter, card_visibility_reference, node_visibility_reference
):
    if isinstance(list_or_dict_to_filter, list):
        return [
            filter_hidden_nodes(
                item, card_visibility_reference, node_visibility_reference
            )
            for item in list_or_dict_to_filter
        ]
    if isinstance(list_or_dict_to_filter, dict):
        dict_to_filter = {**list_or_dict_to_filter}
        for key, val in dict_to_filter.items():
            if isinstance(val, dict) and "@node_id" in val:
                if not card_visibility_reference.get(
                    val["@node_id"], True
                ) or not node_visibility_reference.get(val["@node_id"], True):
                    dict_to_filter[key] = None
                else:
                    dict_to_filter[key] = filter_hidden_nodes(
                        val, card_visibility_reference, node_visibility_reference
                    )
        return dict_to_filter
    raise TypeError


def prepare_links(
    node,
    tile_values,
    node_display_value,
    request_language,
    value_finder,
    is_user_rdm_admin,
):
    links = []

    ### TEMPORARY HELPERS

    def get_resource_labels(tiledata):
        """This is a source of N+1 queries, but we're working around the fact
        that __arches_get_node_display_value() is lossy, i.e. if the display
        values contain the delimiter (", ") we can't distinguish those.
        So we just get the display values again, unfortunately.
        TODO: graduate from the PG function to ORM expressions?
        """
        nonlocal request_language
        ordered_ids = [UUID(innerTileVal["resourceId"]) for innerTileVal in tiledata]
        resources = models.ResourceInstance.objects.filter(pk__in=ordered_ids).in_bulk()
        return [
            (
                resources[res_id]
                .descriptors.get(request_language, {})
                .get("name", _("Undefined"))
                if res_id in resources
                else _("Undefined")
            )
            for res_id in ordered_ids
        ]

    def get_concept_labels(value_ids):
        nonlocal value_finder
        return [
            value_finder.get_value(value_id_str).value for value_id_str in value_ids
        ]

    def get_concept_ids(value_ids):
        nonlocal value_finder
        return [
            value_finder.get_value(value_id_str).concept_id
            for value_id_str in value_ids
        ]

    def form_file_url(tile_url_string):
        prefix = get_script_prefix()
        if tile_url_string.startswith("http") or tile_url_string.startswith(prefix):
            return tile_url_string
        url = get_script_prefix() + tile_url_string
        return url.replace("//", "/")

    ### BEGIN LINK GENERATION
    for tile_val in tile_values:
        if tile_val:
            match node.datatype:
                case "resource-instance":
                    links.append(
                        {
                            "label": node_display_value,
                            "link": get_link(node.datatype, tile_val[0]["resourceId"]),
                        }
                    )
                case "resource-instance-list":
                    labels = get_resource_labels(tile_val)
                    for related_resource, label in zip(tile_val, labels, strict=True):
                        links.append(
                            {
                                "label": label,
                                "link": get_link(
                                    node.datatype, related_resource["resourceId"]
                                ),
                            }
                        )
                case "concept":
                    if is_user_rdm_admin and (
                        concept_id_results := get_concept_ids([tile_val])
                    ):
                        links.append(
                            {
                                "label": node_display_value,
                                "link": get_link(node.datatype, concept_id_results[0]),
                            }
                        )
                case "concept-list":
                    if is_user_rdm_admin:
                        concept_ids = get_concept_ids(tile_val)
                        labels = get_concept_labels(tile_val)
                        for concept_id, label in zip(concept_ids, labels, strict=True):
                            links.append(
                                {
                                    "label": label,
                                    "link": get_link(node.datatype, concept_id),
                                }
                            )
                case "url":
                    links.append(
                        {
                            "label": (
                                tile_val["url_label"]
                                if tile_val["url_label"]
                                else tile_val["url"]
                            ),
                            "link": tile_val["url"],
                        }
                    )
                case "file-list":
                    for file in tile_val:
                        links.append(
                            {
                                "is_file": True,
                                "altText": file.get("altText", {}).get(
                                    request_language
                                )["value"],
                                "attribution": file.get("attribution", {}).get(
                                    request_language
                                )["value"],
                                "title": file.get("title", {}).get(
                                    request_language, ""
                                )["value"],
                                "description": file.get("description", {}).get(
                                    request_language, ""
                                )["value"],
                                "url": form_file_url(file["url"]),
                            }
                        )
                case "reference":
                    for val_id, disp_val in zip(
                        tile_val, json.loads(node_display_value)
                    ):
                        if val_id:
                            links.append(
                                {
                                    "label": disp_val,
                                    "link": get_link(node.datatype, val_id),
                                }
                            )

    return links


def is_number(s):
    try:
        float(s)
        return True
    except:
        return False


def array_from_string(input_str):
    if not input_str:
        return []
    elif is_number(input_str):
        return [input_str]
    else:
        try:
            return json.loads(input_str)
        except json.JSONDecodeError:
            return [input_str]



import copy
import re

from arches.app.permissions.arches_permission_base import (
    get_nodegroups_by_perm_for_user_or_group,
)


def update_report_configuration_with_nodegroup_permissions(
    report_configuration_instance,
    report_nodegroup_ids_with_user_read_permission,
    report_nodegroup_ids_with_user_write_permission,
):
    copy_of_report_configuration = copy.deepcopy(report_configuration_instance.config)

    nodegroup_uuids_by_node_alias = {
        node.alias: node.nodegroup.pk
        for node in report_configuration_instance.graph.node_set.all()
        if node.nodegroup
    }

    def filter_node(node):
        if isinstance(node, dict):
            config = node.get("config", {})

            if isinstance(config, dict):
                nodegroup_alias = config.get("nodegroup_alias")
                nodegroup_id = nodegroup_uuids_by_node_alias.get(nodegroup_alias)

                if (
                    nodegroup_id
                    and nodegroup_id
                    not in report_nodegroup_ids_with_user_read_permission
                ):
                    return None

                config["has_write_permission"] = (
                    nodegroup_id in report_nodegroup_ids_with_user_write_permission
                    if nodegroup_id
                    else False
                )

                for key in ["tabs", "sections", "components", "node_aliases"]:
                    if key in config:
                        filtered = filter_list(config[key])

                        config[key] = filtered

                        looks_like_data_section = "graph_slug" not in config
                        if not filtered and looks_like_data_section:
                            return None

                if descriptor := config.get("descriptor"):
                    config["descriptor"] = filter_descriptor(descriptor)

            if "components" in node:
                filtered = filter_list(node["components"])

                if not filtered:
                    return None

                node["components"] = filtered

            return node

        elif isinstance(node, list):
            return filter_list(node)

        elif isinstance(node, str):
            if (
                (maybe_uuid := nodegroup_uuids_by_node_alias.get(node))
                and maybe_uuid
                and maybe_uuid not in report_nodegroup_ids_with_user_read_permission
            ):
                return None
            else:
                return node

        else:
            return node

    def filter_list(items):
        filtered = []

        for child in items:
            result = filter_node(child)

            if result:
                filtered.append(result)

        return filtered

    def filter_descriptor(descriptor):
        substrings = extract_substrings(descriptor)
        filtered = filter_list(substrings)
        for forbidden in set(substrings) - set(filtered):
            descriptor = descriptor.replace(f"<{forbidden}>", "")
        return descriptor

    return filter_node(copy_of_report_configuration)


def update_report_configuration_for_nodegroup_permissions(
    report_configuration_instance, user
):
    graph_nodegroup_ids = {
        node.nodegroup.pk
        for node in report_configuration_instance.graph.node_set.all()
        if node.nodegroup
    }

    report_nodegroup_ids_with_user_read_permission = set()
    report_nodegroup_ids_with_user_write_permission = set()

    nodegroups_permissions = get_nodegroups_by_perm_for_user_or_group(
        user_or_group=user,
        perms=["models.read_nodegroup", "models.write_nodegroup"],
    )

    for nodegroup, permissions in nodegroups_permissions.items():
        if nodegroup.pk not in graph_nodegroup_ids:
            continue

        if not permissions:  # Empty set implies user has all permissions
            report_nodegroup_ids_with_user_read_permission.add(nodegroup.pk)
            report_nodegroup_ids_with_user_write_permission.add(nodegroup.pk)
            continue

        if "read_nodegroup" in permissions:
            report_nodegroup_ids_with_user_read_permission.add(nodegroup.pk)

        if "write_nodegroup" in permissions:
            report_nodegroup_ids_with_user_write_permission.add(nodegroup.pk)

    return update_report_configuration_with_nodegroup_permissions(
        report_configuration_instance=report_configuration_instance,
        report_nodegroup_ids_with_user_read_permission=report_nodegroup_ids_with_user_read_permission,
        report_nodegroup_ids_with_user_write_permission=report_nodegroup_ids_with_user_write_permission,
    )


def extract_substrings(template_string):
    pattern = r"<(.*?)>"
    substrings = re.findall(pattern, template_string)

    return substrings



import re
from pathlib import Path

from django.core.exceptions import ValidationError
from django.db import models

from arches import VERSION as arches_version
from arches.app.models.models import GraphModel, Node, NodeGroup
from arches.app.models.system_settings import settings
from arches_modular_reports.utils import PrettyJSONEncoder


def get_graph_choices():
    choices = models.Q(isresource=True)
    choices &= ~models.Q(pk=settings.SYSTEM_SETTINGS_RESOURCE_MODEL_ID)
    if arches_version >= (8, 0):
        choices &= models.Q(source_identifier=None)
    return choices


class ReportConfig(models.Model):
    id = models.AutoField(primary_key=True)
    config = models.JSONField(
        blank=True, null=False, default=dict, encoder=PrettyJSONEncoder
    )
    slug = models.TextField(default="default", blank=False, null=False)
    graph = models.ForeignKey(
        GraphModel,
        blank=False,
        on_delete=models.CASCADE,
        related_name="report_configs",
        limit_choices_to=get_graph_choices,
    )

    class Meta:
        managed = True
        db_table = "arches_modular_report_config"
        constraints = [
            models.UniqueConstraint(
                fields=["graph", "slug"],
                name="unique_slug_graph",
            )
        ]

    def __str__(self):
        if self.config and self.graph:
            return f"Config for: {self.graph.name}: {self.config.get('name')}"
        return super().__str__()

    @property
    def excluded_datatypes(self):
        return {"semantic", "annotation", "geojson-feature-collection"}

    def clean(self):
        if not self.graph.slug:
            raise ValidationError("Graph must have a slug")
        if not self.config:
            self.config = self.generate_config()
        self.validate_config()

    def generate_config(self):
        return {
            "name": "Untitled Report",
            "components": [
                {
                    "component": "arches_modular_reports/ModularReport/components/ReportHeader",
                    "config": {
                        "descriptor": f"{self.graph.name} descriptor template",
                        "node_alias_options": {},  # e.g. limit, separator
                    },
                },
                {
                    "component": "arches_modular_reports/ModularReport/components/ReportToolbar",
                    "config": {
                        "lists": True,
                        "export_formats": ["csv", "json-ld", "json"],
                    },
                },
                {
                    "component": "arches_modular_reports/ModularReport/components/ReportTombstone",
                    "config": {
                        "node_aliases": [],
                        "image_node_alias": None,
                        "custom_labels": {},
                    },
                },
                {
                    "component": "arches_modular_reports/ModularReport/components/ReportTabs",
                    "config": {
                        "tabs": [
                            {
                                "name": "Data",
                                "components": [
                                    {
                                        "component": "arches_modular_reports/ModularReport/components/LinkedSections",
                                        "config": {
                                            "sections": self.generate_card_sections()
                                        },
                                    },
                                ],
                            },
                            {
                                "name": "Related Resources",
                                "components": [
                                    {
                                        "component": "arches_modular_reports/ModularReport/components/LinkedSections",
                                        "config": {
                                            "sections": self.generate_related_resources_sections()
                                        },
                                    },
                                ],
                            },
                        ],
                    },
                },
            ],
        }

    def generate_card_sections(self):
        ordered_allowed_nodes = (
            Node.objects.filter(graph=self.graph)
            .exclude(datatype__in=self.excluded_datatypes)
            .exclude(cardxnodexwidget__visible=False)
            .order_by("cardxnodexwidget__sortorder")
        )
        ordered_top_cards = (
            self.graph.cardmodel_set.filter(nodegroup__parentnodegroup__isnull=True)
            .select_related(
                "nodegroup__grouping_node" if arches_version >= (8, 0) else "nodegroup"
            )
            .prefetch_related(
                models.Prefetch(
                    "nodegroup__node_set",
                    ordered_allowed_nodes,
                    to_attr="allowed_nodes",
                )
            )
            .order_by("sortorder")
        )

        def get_grouping_node(nodegroup):
            if arches_version >= (8, 0):
                return nodegroup.grouping_node
            return nodegroup.node_set.filter(pk=nodegroup.pk).first()

        return [
            {
                "name": str(card.name),
                "components": [
                    {
                        "component": "arches_modular_reports/ModularReport/components/DataSection",
                        "config": {
                            "nodegroup_alias": get_grouping_node(card.nodegroup).alias,
                            "node_aliases": [
                                node.alias for node in card.nodegroup.allowed_nodes
                            ],
                            # custom_labels: {node alias: "my custom widget label"}
                            "custom_labels": {},
                            # custom_card_name: "My Custom Card Name"
                            "custom_card_name": None,
                        },
                    }
                ],
            }
            for card in ordered_top_cards
        ]

    def generate_related_resources_sections(self):
        other_graphs = GraphModel.objects.exclude(
            pk=settings.SYSTEM_SETTINGS_RESOURCE_MODEL_ID,
        ).filter(
            slug__isnull=False,
            isresource=True,
        )
        if arches_version >= (8, 0):
            other_graphs = other_graphs.filter(source_identifier=None)
        return [
            {
                "name": str(other_graph.name),
                "components": [
                    {
                        "component": "arches_modular_reports/ModularReport/components/RelatedResourcesSection",
                        "config": {
                            "graph_slug": other_graph.slug,
                            "node_aliases": [],
                            "custom_labels": {},
                        },
                    },
                ],
            }
            for other_graph in other_graphs
        ]

    def validate_config(self):
        def validate_dict(config_dict):
            for key, val in config_dict.items():
                if key == "name":
                    if not isinstance(val, str):
                        raise ValidationError(f"Name is not a string: {val}")
                elif key == "components":
                    if not isinstance(val, list):
                        raise ValidationError(f"Components is not a list: {val}")
                    validate_components(val)
                else:
                    raise ValidationError(f"Invalid key in config: {key}")

        def validate_components(components):
            for item in components:
                for key, val in item.items():
                    if key == "component":
                        if not isinstance(val, str):
                            raise ValidationError(f"Component is not a string: {val}")
                    elif key == "config":
                        if not isinstance(val, dict):
                            raise ValidationError(f"Config is not a dict: {val}")
                        validate_components_config(val)
                    else:
                        raise ValidationError(f"Invalid key in components: {key}")
                component = self.get_or_raise(item, "component", "")
                config = self.get_or_raise(item, "config", "")
                component_name = Path(component).stem.lower()
                method = getattr(self, "validate_" + component_name, lambda _: None)
                # example method: validate_relatedresourcessection
                method(config)

        def validate_components_config(config_dict):
            for val in config_dict.values():
                if isinstance(val, list):
                    for list_item in val:
                        if (
                            isinstance(list_item, dict)
                            and "name" in list_item
                            and "components" in list_item
                        ):
                            validate_dict(list_item)

        validate_dict(self.config)

    def validate_reportheader(self, header_config):
        descriptor_template = self.get_or_raise(header_config, "descriptor", "Header")
        if not isinstance(descriptor_template, str):
            raise ValidationError("Descriptor is not a string")
        substrings = self.extract_node_aliases(descriptor_template)
        usable_nodes = self.graph.node_set.exclude(datatype__in=self.excluded_datatypes)
        self.validate_node_aliases(
            {"node_aliases": substrings},
            "Header",
            usable_nodes,
        )
        if node_alias_options := header_config.get("node_alias_options"):
            self.validate_node_aliases(
                {"node_aliases": node_alias_options.keys()},
                "Header",
                usable_nodes,
            )
            self.validate_options(node_alias_options)

    def validate_reporttombstone(self, tombstone_config):
        self.validate_node_aliases(
            tombstone_config,
            "Tombstone",
            self.graph.node_set.exclude(datatype__in=self.excluded_datatypes),
        )
        if image_node_alias := tombstone_config.get("image_node_alias"):
            if not self.graph.node_set.filter(
                alias=image_node_alias, datatype="file-list"
            ).exists():
                msg = f"Tombstone section contains invalid image node alias: {image_node_alias}"
                raise ValidationError(msg)

    def validate_datasection(self, card_config):
        nodegroup_alias = self.get_or_raise(card_config, "nodegroup_alias", "Data")
        nodegroup = NodeGroup.objects.filter(
            node__alias=nodegroup_alias, node__graph=self.graph
        ).first()
        if not nodegroup:
            raise ValidationError(
                f"Section contains invalid nodegroup: {nodegroup_alias}"
            )

        self.validate_node_aliases(
            card_config,
            "Data",
            nodegroup.node_set.exclude(datatype__in=self.excluded_datatypes),
        )

    def validate_relatedresourcessection(self, rr_config):
        slug = self.get_or_raise(rr_config, "graph_slug", "Related Resources")
        filters = models.Q(slug=slug)
        if arches_version >= (8, 0):
            filters &= models.Q(source_identifier=None)
        try:
            graph = GraphModel.objects.get(filters)
        except (GraphModel.DoesNotExist, GraphModel.MultipleObjectsReturned):
            msg = "Related Resources section contains invalid graph slug"
            raise ValidationError(msg)

        usable_related_nodes = graph.node_set.exclude(
            datatype__in=self.excluded_datatypes
        )
        self.validate_node_aliases(rr_config, "Related Resources", usable_related_nodes)

    def validate_node_aliases(self, config, section_name, usable_nodes_queryset):
        requested_node_aliases = self.get_or_raise(config, "node_aliases", section_name)
        usable_aliases = {node.alias for node in usable_nodes_queryset}
        if extra_node_aliases := set(requested_node_aliases) - usable_aliases:
            raise ValidationError(
                f"{section_name} section contains extraneous "
                "or invalid node aliases or unsupported datatypes: "
                f"{extra_node_aliases}"
            )
        overridden_labels = set(config.get("custom_labels", {}))
        if extra_overridden_labels := overridden_labels - usable_aliases:
            raise ValidationError(
                f"{section_name} section overrides labels for "
                "extraneous or invalid node aliases or unsupported "
                f"datatypes: {extra_overridden_labels}"
            )

    def validate_options(self, options):
        for alias, option_mapping in options.items():
            for key, val in option_mapping.items():
                match key:
                    case "limit":
                        if not isinstance(val, int):
                            raise ValidationError(
                                f"Limit for {alias} is not an integer"
                            )
                    case "separator":
                        if not isinstance(val, str):
                            raise ValidationError(
                                f"Separator for {alias} is not a string"
                            )
                    case _:
                        raise ValidationError(f"Invalid option for {alias}")

    @staticmethod
    def extract_node_aliases(template_string):
        alias_pattern = r"<(.*?)>"
        options_pattern = r"\[(.*?)\]"
        result = []
        for substring in re.findall(alias_pattern, template_string):
            if option_matches := re.findall(options_pattern, substring):
                if len(option_matches) > 1:
                    raise ValidationError("Too many options")
                result.append(re.sub(options_pattern, "", substring, count=1))
            else:
                result.append(substring)
        return result

    @staticmethod
    def get_or_raise(config, key, section_name):
        if key not in config:
            raise ValidationError(f"{section_name} section missing key: {key}")
        return config[key]



# Generated by Django 5.2.5 on 2025-08-28 16:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("arches_modular_reports", "0006_improve_report_config_related_name"),
        ("models", "11499_add_editlog_resourceinstance_idx"),
    ]

    operations = [
        migrations.AddField(
            model_name="reportconfig",
            name="slug",
            field=models.TextField(default="default"),
        ),
        migrations.AddConstraint(
            model_name="reportconfig",
            constraint=models.UniqueConstraint(
                fields=("graph", "slug"), name="unique_slug_graph"
            ),
        ),
    ]



import uuid

from django.db import migrations

template_pk = uuid.UUID("b0908227-ecc2-48dd-931b-314a9031caa0")


class Migration(migrations.Migration):

    dependencies = [
        ("arches_modular_reports", "0001_initial"),
    ]

    def create_template(apps, schema_editor):
        ReportTemplate = apps.get_model("models", "ReportTemplate")
        ReportTemplate(
            pk=template_pk,
            name="Modular Report Template",
            description="A modular report template.",
            component="reports/modular-report",
            componentname="modular-report",
            defaultconfig={},
            preload_resource_data=False,
        ).save()

    def delete_template(apps, schema_editor):
        ReportTemplate = apps.get_model("models", "ReportTemplate")
        ReportTemplate.objects.filter(pk=template_pk).delete()

    operations = [
        migrations.RunPython(create_template, delete_template),
    ]



from django.conf.urls.static import static
from django.conf.urls.i18n import i18n_patterns
from django.urls import include, path

from arches.app.models.system_settings import settings

uuid_regex = settings.UUID_REGEX

urlpatterns = [
    path("", include("arches_modular_reports.urls")),
    path("", include("arches_querysets.urls")),
    path("", include("arches_component_lab.urls")),
]

# handler400 = "arches.app.views.main.custom_400"
# handler403 = "arches.app.views.main.custom_403"
# handler404 = "arches.app.views.main.custom_404"
# handler500 = "arches.app.views.main.custom_500"

# Ensure Arches core urls are superseded by project-level urls
urlpatterns.append(path("", include("arches.urls")))

# Adds URL pattern to serve media files during development
urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

# Only handle i18n routing in active project. This will still handle the routes provided by Arches core and Arches applications,
# but handling i18n routes in multiple places causes application errors.
if settings.ROOT_URLCONF == __name__:
    if settings.SHOW_LANGUAGE_SWITCH is True:
        urlpatterns = i18n_patterns(*urlpatterns)

    urlpatterns.append(path("i18n/", include("django.conf.urls.i18n")))



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("arches_modular_reports", "0002_add_modular_report"),
    ]

    forward_sql_string = """
        CREATE OR REPLACE FUNCTION __arches_get_concept_valueid(
            concept_value UUID,
            language_id   TEXT DEFAULT 'en'
        )
        RETURNS TEXT
        LANGUAGE 'plpgsql'
        COST 100
        VOLATILE PARALLEL UNSAFE
        AS $BODY$

        DECLARE
            concept_id        TEXT := '';
            normalized_lang_id TEXT;
            base_lang_id       TEXT;
        BEGIN
            IF concept_value IS NULL THEN
                RETURN concept_id;
            END IF;

            normalized_lang_id := replace(language_id, '_', '-');
            base_lang_id       := split_part(normalized_lang_id, '-', 1);

            SELECT c.conceptid::text
            INTO concept_id
            FROM values v_orig
            JOIN concepts c   ON v_orig.conceptid = c.conceptid
            JOIN values v_lang 
                ON c.conceptid = v_lang.conceptid
            WHERE v_orig.valueid = concept_value
            AND v_lang.valuetype = 'prefLabel'
            AND (
                v_lang.languageid = normalized_lang_id
                OR v_lang.languageid = base_lang_id
                OR v_lang.languageid LIKE base_lang_id || '-%'
                OR v_lang.languageid LIKE base_lang_id || '_%'
            )
            ORDER BY
            CASE
                WHEN v_lang.languageid = normalized_lang_id THEN 1

                WHEN (
                        normalized_lang_id = base_lang_id 
                        AND 
                        (
                        v_lang.languageid LIKE base_lang_id || '-%' 
                        OR v_lang.languageid LIKE base_lang_id || '_%'
                        )
                    )
                    OR (
                        normalized_lang_id != base_lang_id 
                        AND v_lang.languageid = base_lang_id
                    ) THEN 2

                WHEN v_lang.languageid LIKE base_lang_id || '-%' 
                    OR v_lang.languageid LIKE base_lang_id || '_%' THEN 3

                ELSE 4
            END
            LIMIT 1;

            IF concept_id IS NULL THEN
                concept_id := '';
            END IF;

            RETURN concept_id;
        END;

        $BODY$;


        CREATE OR REPLACE FUNCTION public.__arches_get_concept_list_valueids(
            concept_array jsonb,
            language_id text DEFAULT 'en'
        )
        RETURNS jsonb
        LANGUAGE 'plpgsql'
        COST 100
        VOLATILE PARALLEL UNSAFE

        AS $BODY$
        BEGIN
            IF concept_array IS NULL OR concept_array::text = 'null' THEN
                RETURN '[]'::jsonb;
            END IF;

            RETURN COALESCE(
                (
                    SELECT jsonb_agg(d.valueid)
                    FROM (
                        SELECT __arches_get_concept_valueid(x.conceptid::uuid, language_id) AS valueid
                        FROM (
                            SELECT json_array_elements_text(concept_array::json) AS conceptid
                        ) x
                    ) d
                ),
                '[]'::jsonb 
            );
        END;

        $BODY$;


        CREATE OR REPLACE FUNCTION public.__arches_get_valueid(
            in_tiledata jsonb,
            in_nodeid uuid,
            language_id text DEFAULT 'en'
        )

        RETURNS text
        LANGUAGE 'plpgsql'
        COST 100
        VOLATILE PARALLEL UNSAFE

        AS $BODY$
            declare
                value   text := '';
                in_node_type    text;
                in_node_config  json;
            begin
                if in_nodeid is null or in_nodeid is null then
                    return '<invalid_nodeid>';
                end if;

                if in_tiledata is null then
                    return '';
                end if;

                select n.datatype, n.config
                into in_node_type, in_node_config
                from nodes n where nodeid = in_nodeid::uuid;

                if in_node_type in ('semantic', 'geojson-feature-collection', 'annotation') then
                    return 'unsupported node type (' || in_node_type || ')';
                end if;

                if in_node_type is null then
                    return '';
                end if;

                case in_node_type
                    when 'concept' then
                        value := __arches_get_concept_valueid((in_tiledata ->> in_nodeid::text)::uuid);
                    when 'concept-list' then
                        value := __arches_get_concept_list_valueids(in_tiledata -> in_nodeid::text);
                    when 'resource-instance' then
                        value := __arches_get_resourceinstance_id(in_tiledata -> in_nodeid::text, 'name', language_id);
                    when 'resource-instance-list' then
                        value := __arches_get_resourceinstance_id_list(in_tiledata -> in_nodeid::text, 'name', language_id);
                    when 'url' then
                        value := in_tiledata -> in_nodeid::text ->> 'url';
                    else
                        value := null;
                end case;

                return value;
            end;

                
            $BODY$;


        CREATE OR REPLACE FUNCTION __arches_get_concept_label_v2(
            concept_value UUID,
            language_id TEXT DEFAULT 'en'
        )

        RETURNS TEXT
        LANGUAGE 'plpgsql'
        COST 100
        VOLATILE PARALLEL UNSAFE

        AS $BODY$

        DECLARE
            concept_label      TEXT := '';
            normalized_lang_id TEXT;
            base_lang_id       TEXT;
        BEGIN
            IF concept_value IS NULL THEN
                RETURN concept_label;
            END IF;

            normalized_lang_id := replace(language_id, '_', '-');
            base_lang_id := split_part(normalized_lang_id, '-', 1);

            SELECT v_lang.value
            INTO concept_label
            FROM values v_orig
            JOIN concepts c ON v_orig.conceptid = c.conceptid
            JOIN values v_lang ON c.conceptid = v_lang.conceptid
            WHERE v_orig.valueid = concept_value
            AND v_lang.valuetype = 'prefLabel'
            AND (
                v_lang.languageid = normalized_lang_id
                OR v_lang.languageid = base_lang_id
                OR v_lang.languageid LIKE base_lang_id || '-%'
                OR v_lang.languageid LIKE base_lang_id || '_%'
            )
            ORDER BY
                CASE
                    WHEN v_lang.languageid = normalized_lang_id THEN 1

                    WHEN (normalized_lang_id = base_lang_id AND 
                        (v_lang.languageid LIKE base_lang_id || '-%' OR v_lang.languageid LIKE base_lang_id || '_%'))
                        OR
                        (normalized_lang_id != base_lang_id AND v_lang.languageid = base_lang_id) THEN 2

                    WHEN v_lang.languageid LIKE base_lang_id || '-%' 
                        OR v_lang.languageid LIKE base_lang_id || '_%' THEN 3

                    ELSE 4
                END
            LIMIT 1;

            IF concept_label IS NULL THEN
                concept_label := '';
            END IF;

            RETURN concept_label;
        END;

        $BODY$;

        CREATE OR REPLACE FUNCTION public.__arches_get_concept_list_label_v2(
            concept_array jsonb,
            language_id text DEFAULT 'en'
        )
        RETURNS jsonb  
        LANGUAGE 'plpgsql'
        COST 100
        VOLATILE PARALLEL UNSAFE
        AS $BODY$
        DECLARE
            labels_jsonb jsonb := '[]'::jsonb;
        BEGIN
            IF concept_array IS NULL OR concept_array::text = 'null' THEN
                RETURN labels_jsonb;
            END IF;

            SELECT COALESCE(jsonb_agg(label), '[]'::jsonb)
            INTO labels_jsonb
            FROM (
                SELECT __arches_get_concept_label_v2(conceptid::uuid, language_id) AS label
                FROM json_array_elements_text(concept_array::json) AS elem(conceptid)
            ) sub;

            RETURN labels_jsonb;
        END;
        
        $BODY$;


        CREATE OR REPLACE FUNCTION public.__arches_get_resourceinstance_list_label_v2(
            resourceinstance_value jsonb,
            label_type text DEFAULT 'name'::text,
            language_id text DEFAULT 'en'
        )
        RETURNS jsonb
        LANGUAGE 'plpgsql'
        COST 100
        VOLATILE PARALLEL UNSAFE
        AS $BODY$
        DECLARE
            return_label jsonb := '[]'::jsonb;
        BEGIN
            IF resourceinstance_value IS NULL OR resourceinstance_value::text = 'null' THEN
                RETURN return_label;
            END IF;
            
            SELECT jsonb_agg(dvl.label)
            INTO return_label
            FROM (
                SELECT __arches_get_resourceinstance_label(dv.resource_instance, label_type, language_id) AS label
                FROM (
                    SELECT jsonb_array_elements(resourceinstance_value) AS resource_instance
                ) dv
            ) dvl;
            
            IF return_label IS NULL THEN
                return_label := '[]'::jsonb;
            END IF;
            
            RETURN return_label;
        END;

        $BODY$;

        CREATE OR REPLACE FUNCTION public.__arches_get_resourceinstance_id_list(
            resourceinstance_value jsonb,
            label_type text DEFAULT 'name'::text, 
            language_id text DEFAULT 'en'          
        )
        RETURNS jsonb
        LANGUAGE 'plpgsql'
        COST 100
        VOLATILE PARALLEL UNSAFE
        AS $BODY$
        DECLARE
            return_id_list jsonb := '[]'::jsonb; 
        BEGIN
            IF resourceinstance_value IS NULL OR resourceinstance_value::text = 'null' THEN
                RETURN return_id_list;
            END IF;
            
            SELECT jsonb_agg(dvl.resourceinstance_id)  
            INTO return_id_list
            FROM (
                SELECT public.__arches_get_resourceinstance_id(dv.resource_instance, label_type, language_id) AS resourceinstance_id
                FROM (
                    SELECT jsonb_array_elements(resourceinstance_value) AS resource_instance
                ) dv
            ) dvl;
            
            IF return_id_list IS NULL THEN
                return_id_list := '[]'::jsonb;
            END IF;
            
            RETURN return_id_list;
        END;
        $BODY$;

        CREATE OR REPLACE FUNCTION public.__arches_get_resourceinstance_id(
            resourceinstance_value jsonb,
            label_type text DEFAULT 'name'::text,  
            language_id text DEFAULT 'en'        
        )
        RETURNS text
        LANGUAGE 'plpgsql'
        COST 100
        VOLATILE PARALLEL UNSAFE
        AS $BODY$
        DECLARE
            return_id          text := '';
            target_resourceid  uuid;
        BEGIN

            IF resourceinstance_value IS NULL OR resourceinstance_value::text = 'null' THEN
                RETURN return_id;
            END IF;
            
            target_resourceid := ((resourceinstance_value -> 0) ->> 'resourceId')::uuid;
            
            IF target_resourceid IS NULL THEN
                target_resourceid := (resourceinstance_value ->> 'resourceId')::uuid;
            END IF;
            
            IF target_resourceid IS NULL THEN
                RETURN return_id;
            END IF;
            
            IF EXISTS (
                SELECT 1
                FROM resource_instances r
                WHERE r.resourceinstanceid = target_resourceid
            ) THEN
                return_id := target_resourceid::text;
            ELSE
                return_id := NULL;
            END IF;

            RETURN return_id;
        END;
        $BODY$;

            CREATE OR REPLACE FUNCTION public.__arches_get_node_display_value_v2(
                in_tiledata jsonb,
                in_nodeid uuid,
                language_id text DEFAULT 'en')
                RETURNS text
                LANGUAGE 'plpgsql'
                COST 100
                VOLATILE PARALLEL UNSAFE
            AS $BODY$
                    declare
                        display_value   text := '';
                        in_node_type    text;
                        in_node_config  json;
                    begin
                        if in_nodeid is null or in_nodeid is null then
                            return '<invalid_nodeid>';
                        end if;

                        if in_tiledata is null then
                            return '';
                        end if;

                        select n.datatype, n.config
                        into in_node_type, in_node_config
                        from nodes n where nodeid = in_nodeid::uuid;

                        if in_node_type in ('semantic', 'geojson-feature-collection', 'annotation') then
                            return 'unsupported node type (' || in_node_type || ')';
                        end if;

                        if in_node_type is null then
                            return '';
                        end if;

                        case in_node_type
                            when 'string' then
                                display_value := ((in_tiledata -> in_nodeid::text) -> language_id) ->> 'value';
                            when 'concept' then
                                display_value := __arches_get_concept_label_v2((in_tiledata ->> in_nodeid::text)::uuid);
                            when 'concept-list' then
                                display_value := __arches_get_concept_list_label_v2(in_tiledata -> in_nodeid::text);
                            when 'reference' then
								display_value := __arches_controlled_lists_get_reference_label_list(in_tiledata -> in_nodeid::text, language_id);
                            when 'edtf' then
                                display_value := (in_tiledata ->> in_nodeid::text);
                            when 'file-list' then
                                display_value := __arches_get_file_list_label(in_tiledata -> in_nodeid::text, language_id);
                            when 'domain-value' then
                                display_value := __arches_get_domain_label((in_tiledata ->> in_nodeid::text)::uuid, in_nodeid, language_id);
                            when 'domain-value-list' then
                                display_value := __arches_get_domain_list_label(in_tiledata -> in_nodeid::text, in_nodeid, language_id);
                            when 'url' then
								if length(in_tiledata -> in_nodeid::text ->> 'url_label') > 0 then
									display_value := (in_tiledata -> in_nodeid::text)::jsonb ->> 'url_label';
								else
								 	display_value := (in_tiledata -> in_nodeid::text)::jsonb ->> 'url';
								end if;
                            when 'node-value' then
                                display_value := __arches_get_nodevalue_label(in_tiledata -> in_nodeid::text, in_nodeid);
                            when 'resource-instance' then
                                display_value := __arches_get_resourceinstance_label(in_tiledata -> in_nodeid::text, 'name', language_id);
                            when 'resource-instance-list' then
                                display_value := __arches_get_resourceinstance_list_label_v2(in_tiledata -> in_nodeid::text, 'name', language_id);
                            when 'date' then
                                display_value := TO_CHAR((in_tiledata ->> in_nodeid::text)::timestamp, in_node_config ->> 'dateFormat');
                            else
                                display_value := (in_tiledata ->> in_nodeid::text)::text;

                            end case;

                        return display_value;
                    end;

                        
            $BODY$;
        """

    reverse_sql_string = """
            drop function if exists __arches_get_node_display_value_v2;
            drop function if exists __arches_get_concept_list_label_v2;
            drop function if exists __arches_get_concept_label_v2;
            drop function if exists __arches_get_concept_valueid;
            drop function if exists __arches_get_concept_list_valueids;
            drop function if exists __arches_get_resourceinstance_list_label_v2;
            drop function if exists __arches_get_resourceinstance_id_list;
            drop function if exists __arches_get_resourceinstance_id;
            drop function if exists __arches_get_valueid;
        """

    operations = [
        migrations.RunSQL(forward_sql_string, reverse_sql_string),
    ]



from django.contrib import admin
from django.db import models
from django.forms import Textarea

from arches_modular_reports.models import ReportConfig


class ReportConfigAdmin(admin.ModelAdmin):
    formfield_overrides = {
        models.JSONField: {"widget": Textarea(attrs={"rows": 32, "cols": 100})},
    }


admin.site.register(ReportConfig, ReportConfigAdmin)



from django.urls import include, path, re_path

from arches.app.models.system_settings import settings

from arches_modular_reports.app.views.modular_report import (
    ModularReportAwareResourceReportView,
    NodegroupTileDataView,
    NodePresentationView,
    NodeTileDataView,
    ModularReportConfigView,
    RelatedResourceView,
    UserPermissionsView,
    LanguageSettingsView,
)

uuid_regex = settings.UUID_REGEX

urlpatterns = [
    path(
        "modular_report_config",
        ModularReportConfigView.as_view(),
        name="modular_report_config",
    ),
    # Override core arches resource report view to allow rendering
    # distinct template for modular reports.
    re_path(
        r"^report/(?P<resourceid>%s)$" % uuid_regex,
        ModularReportAwareResourceReportView.as_view(),
        name="resource_report",
    ),
    path(
        "api/related_resources/<uuid:resourceid>/<slug:related_graph_slug>",
        RelatedResourceView.as_view(),
        name="api_related_resources",
    ),
    path(
        "api/node_presentation/<uuid:resourceid>",
        NodePresentationView.as_view(),
        name="api_node_presentation",
    ),
    path(
        "api/nodegroup_tile_data/<uuid:resourceid>/<slug:nodegroup_alias>",
        NodegroupTileDataView.as_view(),
        name="api_nodegroup_tile_data",
    ),
    path(
        "api/node_tile_data/<uuid:resourceid>",
        NodeTileDataView.as_view(),
        name="api_node_tile_data",
    ),
    path(
        "api/has_permissions",
        UserPermissionsView.as_view(),
        name="api_has_permissions",
    ),
    path(
        "api/client_language_settings",
        LanguageSettingsView.as_view(),
        name="api_client_language_settings",
    ),
    path("", include("arches_querysets.urls")),
    path("", include("arches_component_lab.urls")),
]



import functools

from django.core.exceptions import BadRequest, PermissionDenied

from arches.app.models import models
from arches.app.utils.permission_backend import PermissionBackend


def can_read_nodegroup(view_func):
    """
    Decorator to ensure that the user has read permissions for a specific NodeGroup.
    The decorated view must accept 'nodegroup_alias' and 'resourceid' as keyword args.
    """

    @functools.wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        nodegroup_alias = kwargs.get("nodegroup_alias")
        resource_id = kwargs.get("resourceid")
        nodegroup = models.NodeGroup.objects.filter(
            node__graph__resourceinstance=resource_id,
            node__alias=nodegroup_alias,
        ).first()
        if not nodegroup:
            raise BadRequest

        permission_backend = PermissionBackend()

        # TODO: remove superuser check once `has_perm` is updated to handle superusers
        if request.user.is_superuser or permission_backend.has_perm(
            user_obj=request.user, perm="models.read_nodegroup", obj=nodegroup
        ):
            return view_func(request, *args, **kwargs)
        else:
            raise PermissionDenied

    return _wrapped_view



import json
from http import HTTPStatus

from django.core.paginator import Paginator
from django.db.models import Q
from django.http import Http404
from django.shortcuts import render
from django.urls import reverse
from django.utils import translation
from django.utils.decorators import method_decorator
from django.utils.translation import get_language, get_language_info, gettext as _
from django.views.generic import View

from arches import VERSION as arches_version
from arches.app.datatypes.concept_types import BaseConceptDataType
from arches.app.models import models
from arches.app.utils.betterJSONSerializer import JSONDeserializer
from arches.app.utils.decorators import can_read_resource_instance
from arches.app.utils.permission_backend import get_nodegroups_by_perm, group_required
from arches.app.utils.response import JSONErrorResponse, JSONResponse
from arches.app.views.api import APIBase
from arches.app.views.base import MapBaseManagerView
from arches.app.views.resource import ResourceReportView

from arches_modular_reports.app.utils.decorators import can_read_nodegroup
from arches_modular_reports.models import ReportConfig

from arches_modular_reports.app.utils.update_report_configuration_for_nodegroup_permissions import (
    update_report_configuration_for_nodegroup_permissions,
)

from arches_modular_reports.app.utils.nodegroup_tile_data_utils import (
    annotate_node_values,
    annotate_related_graph_nodes_with_widget_labels,
    array_from_string,
    build_valueid_annotation,
    get_sorted_filtered_relations,
    get_sorted_filtered_tiles,
    prepare_links,
)


@method_decorator(can_read_resource_instance, name="dispatch")
class ModularReportConfigView(View):
    def get(self, request):
        filters = Q(graph__resourceinstance=request.GET.get("resourceId"))

        if arches_version >= (8, 0):
            filters &= Q(graph__source_identifier=None)

        slug = request.GET.get("slug", None)
        if slug:
            filters &= Q(slug__iexact=slug)
        else:
            filters &= Q(slug__iexact="default")

        config_instance = (
            ReportConfig.objects.select_related("graph")
            .prefetch_related("graph__node_set", "graph__node_set__nodegroup")
            .get(filters)
        )

        if not config_instance:
            return JSONErrorResponse(
                _("No report config found."), status=HTTPStatus.NOT_FOUND
            )

        return JSONResponse(
            update_report_configuration_for_nodegroup_permissions(
                config_instance, request.user
            )
        )


@method_decorator(can_read_resource_instance, name="dispatch")
class ModularReportAwareResourceReportView(ResourceReportView):
    def get(self, request, resourceid=None):
        graph = (
            models.GraphModel.objects.filter(resourceinstance=resourceid)
            .select_related("template")
            .first()
        )
        if not graph:
            raise Http404(
                _("No active report template is available for this resource.")
            )

        if graph.template.componentname == "modular-report":
            template = "views/resource/modular_report.htm"
            # Skip a few queries by jumping over the MapBaseManagerView
            # and calling its parent. This report doesn't use a map.
            context = super(MapBaseManagerView, self).get_context_data(
                main_script="views/resource/report",
                resourceid=resourceid,
                templateid=graph.template.pk,
                graph_slug=graph.slug,
                # To the extent possible, avoid DB queries needed for KO
                report_templates=[graph.template],
                card_components=models.CardComponent.objects.none(),
                widgets=models.Widget.objects.none(),
                map_markers=models.MapMarker.objects.none(),
                geocoding_providers=models.Geocoder.objects.none(),
            )
        else:
            name_resource = (
                models.ResourceInstance.objects.only("name")
                .get(resourceinstanceid=str(resourceid))
                .name
            )
            template = "views/resource/report.htm"
            context = self.get_context_data(
                main_script="views/resource/report",
                resourceid=resourceid,
                report_templates=models.ReportTemplate.objects.all(),
                card_components=models.CardComponent.objects.all(),
                widgets=models.Widget.objects.all(),
                map_markers=models.MapMarker.objects.all(),
                geocoding_providers=models.Geocoder.objects.all(),
                graph_name=graph.name,
                name_resource=name_resource,
            )

        if graph.iconclass:
            context["nav"]["icon"] = graph.iconclass
        context["nav"]["title"] = graph.name
        context["nav"]["res_edit"] = True
        context["nav"]["print"] = True

        return render(request, template, context)


@method_decorator(can_read_resource_instance, name="dispatch")
class RelatedResourceView(APIBase):
    def get(self, request, resourceid, related_graph_slug):
        try:
            resource = models.ResourceInstance.objects.get(pk=resourceid)
            filters = Q(slug=related_graph_slug)
            if arches_version >= (8, 0):
                filters &= Q(source_identifier=None)
            related_graph = models.GraphModel.objects.get(filters)
        except (models.ResourceInstance.DoesNotExist, models.GraphModel.DoesNotExist):
            return JSONErrorResponse(status=HTTPStatus.NOT_FOUND)

        additional_nodes = request.GET.get("node_aliases", "").split(",")
        page_number = request.GET.get("page", 1)
        rows_per_page = request.GET.get("rows_per_page", 10)
        sort_field = request.GET.get("sort_field", "@relation_name")
        direction = request.GET.get("direction", "asc")
        query = request.GET.get("query", "")
        request_language = translation.get_language()

        permitted_nodegroups = get_nodegroups_by_perm(
            request.user, "models.read_nodegroup"
        )
        is_user_rdm_admin = group_required(request.user, "RDM Administrator")

        nodes = annotate_related_graph_nodes_with_widget_labels(
            additional_nodes, related_graph, request_language
        )
        relations = get_sorted_filtered_relations(
            resource=resource,
            related_graph=related_graph,
            nodes=nodes,
            permitted_nodegroups=permitted_nodegroups,
            sort_field=sort_field,
            direction=direction,
            query=query,
            request_language=request_language,
        )
        paginator = Paginator(relations, rows_per_page)
        result_page = paginator.get_page(page_number)

        def make_resource_report_link(relation):
            nonlocal resourceid
            # Both sides are UUID python types (from ORM, or from route)
            if arches_version < (8, 0):
                if relation.resourceinstanceidfrom_id == resourceid:
                    target = relation.resourceinstanceidto_id
                else:
                    target = relation.resourceinstanceidfrom_id
            else:
                if relation.from_resource_id == resourceid:
                    target = relation.to_resource_id
                else:
                    target = relation.from_resource_id
            return reverse("resource_report", args=[target])

        value_finder = BaseConceptDataType()  # fetches serially, but caches.

        response_data = {
            "results": [
                {
                    "@relation_name": {
                        "display_value": getattr(relation, "@relation_name"),
                        "links": [],
                    },
                    "@display_name": {
                        "display_value": getattr(relation, "@display_name"),
                        "links": [
                            {
                                "label": getattr(relation, "@display_name"),
                                "link": make_resource_report_link(relation),
                            }
                        ],
                    },
                    **{
                        node.alias: {
                            "display_value": getattr(relation, node.alias),
                            "links": prepare_links(
                                node=node,
                                tile_values=getattr(
                                    relation, node.alias + "_instance_details", []
                                ),
                                node_display_value=getattr(relation, node.alias),
                                request_language=request_language,
                                value_finder=value_finder,
                                is_user_rdm_admin=is_user_rdm_admin,
                            ),
                        }
                        for node in nodes
                    },
                }
                for relation in result_page
            ],
            "graph_name": related_graph.name,
            "widget_labels": {node.alias: node.widget_label for node in nodes},
            "total_count": paginator.count,
            "page": result_page.number,
        }

        return JSONResponse(response_data)


class NodePresentationView(APIBase):
    @method_decorator(can_read_resource_instance, name="dispatch")
    def get(self, request, resourceid):
        try:
            graph = models.GraphModel.objects.filter(resourceinstance=resourceid).get()
        except models.GraphModel.DoesNotExist:
            return JSONErrorResponse(status=HTTPStatus.NOT_FOUND)
        permitted_nodegroups = get_nodegroups_by_perm(
            request.user, "models.read_nodegroup"
        )
        nodes = (
            models.Node.objects.filter(graph=graph)
            .filter(nodegroup__in=permitted_nodegroups)
            .select_related("nodegroup")
            .prefetch_related(
                "nodegroup__cardmodel_set",
                "cardxnodexwidget_set__widget",
            )
        )

        def getattr_from_queryset(queryset, attr, fallback):
            if queryset:
                return getattr(queryset[0], attr, fallback)
            return fallback

        def get_widget_name(queryset, fallback):
            if queryset and queryset[0].widget:
                return getattr(queryset[0].widget, "name", fallback)
            return fallback

        def get_widget_format(queryset, fallback=None):
            if fallback is None:
                fallback = {"format": "", "prefix": {}, "suffix": {}}
            if queryset and queryset[0].widget:
                if getattr(queryset[0].widget, "name", None) != "number-widget":
                    return fallback
                config = getattr(queryset[0], "config", None)
                if config:
                    ret = {
                        "format": config.get("format", fallback),
                        "prefix": config.get("prefix", fallback),
                        "suffix": config.get("suffix", fallback),
                    }
                    return ret
            return fallback

        def get_node_visibility(node):
            if node.pk == node.nodegroup.pk and node.nodegroup.cardmodel_set.all():
                return node.nodegroup.cardmodel_set.all()[0].visible
            if node.cardxnodexwidget_set.all():
                return node.cardxnodexwidget_set.all()[0].visible
            return True

        return JSONResponse(
            {
                node.alias: {
                    "nodeid": node.nodeid,
                    "name": node.name,
                    "card_name": getattr_from_queryset(
                        node.nodegroup.cardmodel_set.all(),
                        "name",
                        "",
                    ),
                    "card_order": getattr_from_queryset(
                        node.nodegroup.cardmodel_set.all(),
                        "sortorder",
                        0,
                    ),
                    "widget_label": getattr_from_queryset(
                        node.cardxnodexwidget_set.all(),
                        "label",
                        node.name.replace("_", " ").title(),
                    ),
                    "widget_order": getattr_from_queryset(
                        node.cardxnodexwidget_set.all(),
                        "sortorder",
                        0,
                    ),
                    "visible": get_node_visibility(node),
                    "nodegroup": {
                        "nodegroup_id": node.nodegroup.pk,
                        "cardinality": node.nodegroup.cardinality,
                    },
                    "is_rich_text": get_widget_name(
                        node.cardxnodexwidget_set.all(), None
                    )
                    == "rich-text-widget",
                    "is_numeric": get_widget_name(node.cardxnodexwidget_set.all(), None)
                    == "number-widget",
                    "number_format": get_widget_format(node.cardxnodexwidget_set.all()),
                }
                for node in nodes
            }
        )


@method_decorator(can_read_resource_instance, name="dispatch")
@method_decorator(can_read_nodegroup, name="dispatch")
class NodegroupTileDataView(APIBase):
    def get(self, request, resourceid, nodegroup_alias):
        page_number = request.GET.get("page")
        rows_per_page = request.GET.get("rows_per_page")
        filters = request.GET.get("filters", None)
        query = request.GET.get("query")
        sort_node_id = request.GET.get("sort_node_id")
        direction = request.GET.get("direction", "asc")

        filters = JSONDeserializer().deserialize(filters) if filters else None

        user_language = translation.get_language()

        is_user_rdm_admin = group_required(request.user, "RDM Administrator")

        tiles = get_sorted_filtered_tiles(
            resourceinstanceid=resourceid,
            nodegroup_alias=nodegroup_alias,
            sort_node_id=sort_node_id,
            direction=direction,
            query=query,
            user_language=user_language,
            user=request.user,
            filters=filters,
        )

        paginator = Paginator(tiles, rows_per_page)
        page = paginator.page(page_number)

        response_data = {
            "results": [
                {
                    **{
                        key: build_valueid_annotation(
                            value, is_user_rdm_admin, user_language
                        )
                        for key, value in tile.alias_annotations.items()
                    },
                    "@has_children": tile.has_children,
                    "@tile_id": tile.tileid,
                }
                for tile in page.object_list
            ],
            "total_count": paginator.count,
            "page": page.number,
        }

        return JSONResponse(response_data)


@method_decorator(can_read_resource_instance, name="dispatch")
class NodeTileDataView(APIBase):
    def get(self, request, resourceid):
        permitted_nodegroups = get_nodegroups_by_perm(request.user, "read_nodegroup")
        node_aliases = request.GET.getlist("node_alias", [])
        user_lang = translation.get_language()
        tile_limit = int(request.GET.get("tile_limit", 0))

        is_user_rdm_admin = group_required(request.user, "RDM Administrator")

        nodes_with_display_data = annotate_node_values(
            node_aliases, resourceid, permitted_nodegroups, user_lang, tile_limit
        )
        value_finder = BaseConceptDataType()

        return JSONResponse(
            {
                node.alias: [
                    {
                        "display_values": array_from_string(
                            display_object["display_value"]
                        ),
                        "links": prepare_links(
                            node,
                            [display_object["tile_value"]],
                            display_object["display_value"],
                            user_lang,
                            value_finder,
                            is_user_rdm_admin,
                        ),
                    }
                    for display_object in node.display_data
                ]
                for node in nodes_with_display_data
            }
        )


class UserPermissionsView(APIBase):
    def get(self, request):
        reqested_permissions = json.loads(request.GET.get("permissions", "[]"))
        user_permissions = {}
        for permission in reqested_permissions:
            if permission == "RDM Administrator":
                user_permissions[permission] = group_required(request.user, permission)
        return JSONResponse(user_permissions)


class LanguageSettingsView(APIBase):
    def get(self, request):
        return JSONResponse(
            {
                "language": get_language(),
                "language_dir": (
                    "ltr" if not get_language_info(get_language())["bidi"] else "rtl"
                ),
            }
        )
